import json
import math
import os
from typing import Union

import pygame


class TileMap:
    def __init__(self, dct=None):
        self.tiles = dct if dct else {}

    def tuplify(self):
        tmp = {}
        for i, v in self.tiles.items():
            tmp[parse_map_key(i)] = v
        self.tiles = tmp

    def stringify(self):
        tmp = {}
        for i, v in self.tiles.items():
            tmp[format_map_key(i)] = v

    def get_tile(self, tile_pos: Union[pygame.Vector2, tuple[int, int]]):
        return self.tiles[tile_pos[0],tile_pos[1]]

    # Set tile
    def set_tile(self, tile_pos: Union[pygame.Vector2, tuple[int, int]], value: str) -> None:
        self.tiles[tile_pos[0], tile_pos[1]] = value

    # If chunk is in map
    def __contains__(self, tile_pos: Union[pygame.Vector2, tuple[int, int]]) -> bool:
        return (tile_pos[0], tile_pos[1]) in self.tiles

    # Delete tile from map. If containing chunk is now empty, delete that chunk
    def del_tile(self, tile_pos: Union[pygame.Vector2, tuple[int, int]]) -> None:
        del self.tiles[tile_pos[0], tile_pos[1]]

    def get_visible_tiles(self, cam):
        mn, mx = cam.bounds
        chunks = {}
        for x in range(int(mn.x), int(mx.x) + 1):
            for y in range(int(mn.y), int(mx.y) + 1):
                if self.__contains__((x, y)):
                    chunks[format_map_key((x, y))] = self.get_tile((x,y))
        return chunks

    @staticmethod
    def load_map(fn: str):
        if os.path.exists(fn):
            with open(fn, "r+") as file:
                data = json.loads(file.read())
                t = TileMap(data)
                t.tuplify()
        else:
            raise FileNotFoundError(f"Path '{fn}' not found")

    def save(self, fn: str):
        self.tuplify()
        with open(fn, "w+") as file:
            file.write((json.dumps(self.dict)))



def parse_map_key(key: str):
    return tuple(int(i) for i in key.split(","))


def format_map_key(loc: Union[tuple[int, int], pygame.Vector2]):
    return f"{int(loc[0])},{int(loc[1])}"



from typing import Union

import pygame

class Camera:
    # TODO: 0,0 should be at center
    def __init__(self, screen_size, viewport_size):
        self.v_size: pygame.Vector2 = pygame.Vector2(viewport_size)
        self.sc_size: pygame.Vector2 = pygame.Vector2(screen_size)
        self.pos: pygame.Vector2 = pygame.Vector2(0, 0)
        self.scale: pygame.Vector2 = pygame.Vector2()
        self._update_scale()

    def project_x(self, x: Union[int, float]):
        return (x - self.pos.x + self.center.x) * self.scale.x

    def project_y(self, y: Union[int, float]):
        return (y - self.pos.y + self.center.y) * self.scale.y

    def unproject_x(self, x: Union[int, float]):
        return x / self.scale.x + self.pos.x - self.center.x

    def unproject_y(self, y: Union[int, float]):
        return y / self.scale.y + self.pos.y - self.center.y

    def project_x_dist(self, x: Union[int, float]):
        return x * self.scale.x

    def project_y_dist(self, y: Union[int, float]):
        return y * self.scale.y

    def unproject_x_dist(self, x: Union[int, float]):
        return x / self.scale.x

    def unproject_y_dist(self, y: Union[int, float]):
        return y / self.scale.y

    def project(self, pos: Union[tuple[int, int], tuple[float, float]]) -> tuple:
        return self.project_x(pos[0]), self.project_y(pos[1])

    def unproject(self, pos: Union[tuple[int, int], tuple[float, float]]) -> tuple:
        return self.unproject_x(pos[0]), self.unproject_y(pos[1])

    def project_dist(self, size: Union[tuple[int, int], tuple[float, float], pygame.Vector2]) -> tuple:
        return self.project_x_dist(size[0]), self.project_y_dist(size[1])

    def unproject_dist(self, size: Union[tuple[int, int], tuple[float, float], pygame.Vector2]) -> tuple:
        return self.unproject_x_dist(size[0]), self.unproject_y_dist(size[1])

    def project_rect(self,
                     rect: Union[tuple[int, int, int, int], tuple[float, float, float, float], tuple[
                         tuple[int, int], tuple[int, int]], tuple[
                                     tuple[float, float], tuple[float, float]], pygame.Rect]):
        if len(rect) == 4 and type(rect) == tuple:
            return *self.project((rect[0], rect[1])), *self.project_dist((rect[2], rect[3]))
        elif len(rect) == 2:
            return self.project(rect[0]), self.project_dist(rect[1])
        elif type(rect) == pygame.Rect:
            return pygame.Rect((self.project_x(rect.x), self.project_y(rect.y), self.project_x_dist(rect.width),
                                self.project_y_dist(rect.height)))

    def unproject_rect(self,
                       rect: Union[tuple[int, int, int, int], tuple[float, float, float, float], tuple[
                           tuple[int, int], tuple[int, int]], tuple[
                                       tuple[float, float], tuple[float, float]], pygame.Rect]):
        if len(rect) == 4 and type(rect) == tuple:
            return *self.unproject((rect[0], rect[1])), *self.unproject_dist((rect[2], rect[3]))
        elif len(rect) == 2:
            return self.unproject(rect[0]), self.unproject_dist(rect[1])
        elif type(rect) == pygame.Rect:
            return pygame.Rect((self.unproject_x(rect.x), self.unproject_y(rect.y), self.unproject_x_dist(rect.width),
                                self.unproject_y_dist(rect.height)))

    def translate(self, pos: Union[tuple[int, int], tuple[float, float], pygame.Vector2]):
        self.pos += pos

    def set_scale(self, scale: Union[tuple[int, int], tuple[float, float], pygame.Vector2]):
        if type(scale) is tuple:
            self.scale.update(scale)
        else:
            self.scale = scale
        self.v_size.update(self.sc_size.x / self.scale.x, self.sc_size.y / self.scale.y)

    def _update_scale(self):
        self.scale.update(self.sc_size.x / self.v_size.x, self.sc_size.y / self.v_size.y)

    def resize_viewport(self, size: Union[tuple[int, int], pygame.Vector2]):
        if type(size) is tuple:
            self.v_size.update(size)
        else:
            self.v_size = size
        self._update_scale()

    def resize_screen(self, size: Union[tuple[int, int], pygame.Vector2]):
        if type(size) is tuple:
            self.sc_size.update(size)
        else:
            self.sc_size = size
        self._update_scale()

    @property
    def bounds(self):
        return pygame.Vector2(self.pos.x - self.center.x, self.pos.y - self.center.y), pygame.Vector2(
            self.pos.x + self.center.x, self.pos.y + self.center.y)

    @property
    def center(self):
        return self.v_size/2



import colorsys
import random


class Color:
    def __init__(self, r: int, g: int, b: int, a: int = 255):
        self.r = r
        self.g = g
        self.b = b
        self.a = a

    @staticmethod
    def random_rgb():
        return Color(random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))

    @staticmethod
    def from_hsv(h, s, v):
        c = colorsys.hsv_to_rgb(h / 360, s / 255, v / 255)
        print(c)
        return Color(round(c[0] * 255), round(c[1] * 255), round(c[2] * 255))

    @staticmethod
    def from_hex(h: str):
        h = h.strip("#")
        if len(h) == 6:
            return Color(int(h[:2], 16), int(h[2:4], 16), int(h[4:],16))
        elif len(h) == 8:
            return Color(int(h[:2], 16), int(h[2:4], 16), int(h[4:6], 16), int(h[6:]))
        else:
            raise ValueError("Hex string must be 7 or 9 characters long (including '#')")

    @property
    def as_hex(self):
        return ('#%02x%02x%02x' % (self.r, self.g, self.b)) if self.a == 255 else '#%02x%02x%02x%02x' % (
            self.r, self.g, self.b, self.a)

    @property
    def as_hsv(self):
        c = colorsys.rgb_to_hsv(self.r / 255, self.g / 255, self.b / 255)
        return round(c[0] * 360), round(c[1] * 255), round(c[2] * 255)

    @property
    def as_tuple(self):
        if self.a == 255:
            return self.r, self.g, self.b
        return self.r, self.g, self.b, self.a

    def __repr__(self):
        s = f"[{self.r}, {self.g}, {self.b}"
        return s + f", {self.a}]" if self.a != 255 else s + "]"



import os
from typing import Union

import pygame


def clamp(v: Union[int, float], mn: Union[int, float], mx: Union[int, float]):
    return max(min(v, mx), mn)


def img_clip(img: pygame.Surface, rect: tuple[int, int, int, int]):
    return img.subsurface(rect)



import os

import pygame


def load_dir(directory: str):
    imgs = {}
    d = os.fsencode(directory)
    for file in os.listdir(d):
        fn = os.fsdecode(file)
        if fn.endswith(".jpg") or fn.endswith(".png") or fn.endswith("jpeg"):
            imgs[os.path.splitext(fn)[0]] = load(f"./{directory}/{fn}")
    return imgs


def load(path: str):
    return pygame.image.load(path).convert()


def load_all(directory:str):
    imgs = {}
    files = [entry.path.strip("./") for entry in [f for f in os.scandir(directory)] if entry.is_file()]
    for i in files:
        if i.endswith(".jpg") or i.endswith(".png") or i.endswith("jpeg"):
            imgs[os.path.splitext(i)[0]] = load(i)
    return imgs
